# This file is a part of sedbgmux, an open source DebugMux client.
# Copyright (c) 2022  Vadim Yanitskiy <fixeria@osmocom.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import logging
import threading
import queue
import sys

from . import DbgMuxConnHandler
from . import DbgMuxConnState

# local logger for this module
log = logging.getLogger(__name__)


class DbgMuxConnTerminal(DbgMuxConnHandler):
    ''' Terminal for communicating with 'Interactive Debug' DPs '''

    def __init__(self):
        super().__init__()

        self.attached: bool = False
        self._rx_thread = threading.Thread(target=self._rx_worker,
                                           daemon=True)
        self._shutdown = threading.Event()

    def attach(self) -> None:
        ''' Read data blocks from stdin and enqueue to the Tx queue '''
        self.attached = True
        while not self._shutdown.is_set():
            try:
                line: str = input()
                if self.conn_state == DbgMuxConnState.Established:
                    self.send_data(bytes(line + '\r', 'ascii'))
            except (KeyboardInterrupt, EOFError) as e:
                break
        self.attached = False

    def _rx_worker(self) -> None:
        ''' Dequeue data blocks from the Rx queue and print to stdout '''
        while not self._shutdown.is_set():
            try:
                data: bytes = self._rx_data_queue.get(block=True, timeout=0.2)
                if self.attached:
                    sys.stdout.write(data.decode('ascii'))
                    sys.stdout.flush()
                self._rx_data_queue.task_done()
            except queue.Empty:
                pass
        log.debug('Thread \'%s-Rx\' is shutting down', self.__class__.__name__)

    def _conn_established(self) -> None:
        ''' Called when a connection has been established '''
        self._shutdown.clear()
        self._rx_thread.start()

    def _conn_terminated(self) -> None:
        ''' Called when a connection has been terminated '''
        self._shutdown.set()
        self._rx_thread.join()
